#include once "mdList.bi"
#include once "fbgfx.bi"

namespace GuiState
  const active = 0, deactivated = 1, activated = 2, noBorder = 4, lightBorder = 8, focus = 16, hidden = 128
end namespace

enum GuiElementType
  button = 1, check, radio, label, textfield
end enum

type TGuiElement
  as long id
  as GuiElementType etype
  as integer x, y, w, h, fontsize, state, ctrl
  as string title, key, info
  as TExercise ptr exercPtr
  as FB.Image ptr img
  declare constructor
  declare constructor(etype as GuiElementType, id as integer = 0, _
                      x as integer, y as integer, w as integer = 0, h as integer = 0, title as string = "", _
                      key as string = "", ctrl as integer = 0, _
                      info as string = "", state as integer = GuiState.active, _
                      size as integer = 13)
  declare constructor(etype as GuiElementType, id as integer = 0, _
                      imgStr as string, x as integer, y as integer, w as integer = 0, h as integer = 0, _
                      key as string = "", ctrl as integer = 0, _
                      info as string = "", state as integer = GuiState.active, _
                      size as integer = 13)
  declare constructor (etype as GuiElementType, id as integer = 0, _
                       x as integer, y as integer, w as integer = 0, h as integer = 0, exerc as TExercise ptr, _
                       key as string = "", ctrl as integer = 0, _
                       info as string = "", state as integer = GuiState.active, _
                       size as integer = 13)
  declare operator cast as string
  declare sub show(state as integer = GuiState.active)
end type

constructor TGuiElement
end constructor

constructor TGuiElement(etype as GuiElementType, id as integer = 0, _
                        x as integer, y as integer, w as integer = 0, h as integer = 0, title as string = "", _
                        key as string = "", ctrl as integer = 0, _
                        info as string = "", state as integer = GuiState.active, _
                        size as integer = 13)
  this.eType    = etype
  this.id       = id
  this.x        = x
  this.y        = y
  this.w        = w
  this.h        = h
  this.state    = state
  this.title    = title
  this.key      = key
  this.ctrl     = ctrl
  this.info     = info
  this.fontsize = size
end constructor

constructor TGuiElement(etype as GuiElementType, id as integer = 0, _
                        imgStr as string, x as integer, y as integer, w as integer = 0, h as integer = 0, _
                        key as string = "", ctrl as integer = 0, _
                        info as string = "", state as integer = GuiState.active, _
                        size as integer = 13)
  this = TGuiElement(etype, id, x, y, w, h, "", key, ctrl, info, state, size)
  dim as integer imgW, imgH, filenr = freefile
  if open(imgStr for binary as #filenr) = 0 then
    if lof(filenr) >= 32 then
      get #filenr, 19, imgW
      get #filenr, 23, imgH
      close #filenr
      this.img = imagecreate(imgW, imgH)
      bload imgStr, this.img
    else
      close #filenr
    end if
  end if
end constructor

constructor TGuiElement(etype as GuiElementType, id as integer = 0, _
                        x as integer, y as integer, w as integer = 0, h as integer = 0, exerc as TExercise ptr, _
                        key as string = "", ctrl as integer = 0, _
                        info as string = "", state as integer = GuiState.active, _
                        size as integer = 13)
  this = TGuiElement(etype, id, x, y, w, h, "", key, ctrl, info, state, size)
  exercPtr = exerc
end constructor

operator TGuiElement.cast as string
  return iif(@exercPtr andalso @(exercPtr->info), (exercPtr->info)(*exercPtr), title)
end operator

operator =(e1 as TGuiElement, e2 as TGuiElement) as integer
  return e1.id = e2.id
end operator

sub TGuiElement.show(stateVar as integer = GuiState.active)
  if state and GuiState.hidden then exit sub
  dim as uinteger col = iif(state and GuiState.deactivated orelse stateVar and GuiState.deactivated, &h999999, 0)
  select case eType
    case GuiElementType.button, GuiElementType.check, GuiElementType.label
      if state and GuiState.lightBorder then
        line (x, y)-step(w+1, h+1), col, b
      elseif (state and GuiState.noBorder) = 0 then
        line (x, y)-step(w, h), col, bf
      end if
      if (state and GuiState.activated) = 0 then line (x+2, y+2)-step(w-4, h-4), rgb(255, 255, 255), bf
    case GuiElementType.radio
      dim as integer radius = iif(w > h, w\2, h\2)
      circle (x+w\2, y+h\2), radius,,,, h/w
      if (state and GuiState.activated) then circle (x+w\2, y+h\2), radius-2,,,, h/w, f
  end select
  if img then
    dim as integer x_, y_, w_, h_
    if w > img->width + 4 then
      w_ = img->width
      x_ = (w-img->width)\2
    else
      w_ = w - 4
      x_ = 2
    end if
    if h > img->height + 4 then
      h_ = img->height
      y_ = (h-img->height)\2
    else
      h_ = h - 4
      y_ = 2
    end if
    if state and GuiState.activated then
      put (x+x_, y+y_), img, ((img->width-w_)\2, (img->height-h_)\2)-step(w_-1, h_-1), preset
    else
      put (x+x_, y+y_), img, ((img->width-w_)\2, (img->height-h_)\2)-step(w_-1, h_-1), pset
    end if
    if state and GuiState.deactivated then
      dim as FB.Image ptr tempImg = imageCreate(w, h, rgb(95, 95, 95))
      put (x, y), tempImg, add
    end if
  else
    dim as string txt = iif(exercPtr, exercPtr->info(*exercPtr), title)
    if (state and (GuiState.deactivated or GuiState.activated)) = (GuiState.deactivated or GuiState.activated) then
      if w > 0 then
        dim as FB.Image ptr tempImg = imageCreate(w-4, h-4, col)
        centerFT tempImg, txt, font, fontsize, col
        put (x+2, y+2), tempImg, pset
        imagedestroy tempImg
      else
        printFT x, y, txt, font, fontSize, col
      end if
    else
      if w > 0 then
        dim as FB.Image ptr tempImg = imageCreate(w-4, h-4, rgb(255, 255, 255))
        centerFT tempImg, txt, font, fontsize, col
        if state and GuiState.activated then
          put (x+2, y+2), tempImg, preset
        else
          put (x+2, y+2), tempImg, pset
        end if
        imagedestroy tempImg
      else
        printFT x, y, txt, font, fontSize, col
      end if
    end if
  end if
  if state and GuiState.focus then line (x+2, y+2)-step(w-4, h-4), 0, b
  if len(info) then
    dim as TCoord coord = FTsize(info, font, 12)
    line (x+w\2-coord.x\2-1, y+h+15-coord.y\2-1)-step(coord.x+2, coord.y+2), &hffffff, bf 
    centerFT x+w\2, y+h+15, info, font, 12, col
  end if
end sub

mdListDeclare(TGuiElement)

type TGuiEvent
  as integer id, mButton
  as string key
end type

type TGui
  as mdList(TGuiElement) elements
  as integer lockedID
  declare destructor
  declare sub show(state as integer = GuiState.active)
  declare function addElement overload(element as TGuiElement) as integer
  declare function getEvent(key as string = "", lock_ as integer = -1) as TGuiEvent
  declare function getElement(id as integer) as TGuiElement
  declare sub setElement(element as TGuiElement, addIfMissing as integer = 0)
end type

destructor TGui
  dim as TGuiElement e
  foreach(TGuiElement, e, elements)
    if e.img then imagedestroy e.img
  nexteach
  elements.clear
end destructor

function TGui.addElement(element as TGuiElement) as integer
  dim as TGuiElement e
  if element.id = 0 then
    ' create id automatically
    if elements.size = 0 then
      element.id = 1
      elements.add(element)
    else
      dim as integer maximum = 0
      foreach(TGuiElement, e, elements)
        if maximum < e.id then maximum = e.id
      nexteach
      element.id = maximum + 1
      elements.add(element)
    end if
  elseif element.id < 0 then
    ' id doesn't matter (e. g. labels)
    ' It's possible to create more elements with the same negative number
    elements.add(element)    
  else
    foreach(TGuiElement, e, elements)
      if e.id = element.id then return 0
    nexteach
    elements.add(element)
  end if
  return element.id
end function

function TGui.getEvent(key as string = "", lock_ as integer = -1) as TGuiEvent
  ' check GUI event (GUI button pressed etc.)
  ' lock_ <> 0: lock an event to prevent the return of the same mouse event two times
  dim as TGuiEvent ret
  dim as TGuiElement e
  dim as integer mx, my, mb, s
  getmouse mx, my,, mb
  foreach(TGuiElement, e, elements)
    s = e.state and not (GuiState.activated or GuiState.noBorder or GuiState.lightBorder or GuiState.focus)
    if s = GuiState.active then
      if key <> "" andalso key = e.key then
        #ifdef checkCtrlAltShift
          if checkCtrlAltShift(e.ctrl) = 0 then continue while
        #endif
        ret.id = e.id
        ret.mButton = -1
        ret.key = key
        exit while
      end if
      if mb > 0 andalso checkMouseIn(e.x, e.y, e.w, e.h) <> 0 then
        if lock_ then
          ret.id = iif(e.id = lockedID, 0, e.id)
          lockedID = e.id
        else
          ret.id = e.id
          lockedID = 0
        end if
        ret.mButton = mb
        ret.key = ""
        exit while
      elseif mb = 0 then
        lockedID = 0
        ret.id = 0
      end if
    end if
  nexteach
  return ret
end function

function TGui.getElement(id as integer) as TGuiElement
  dim as TGuiElement e
  foreach(TGuiElement, e, elements)
    if e.id = id then return e
  nexteach
end function

sub TGui.setElement(element as TGuiElement, addIfMissing as integer = 0)
  dim as TGuiElement e
  dim as integer i = 0
  foreach(TGuiElement, e, elements)
    if e.id = element.id then
      elements.add(i, element)
      elements.remove(i+1)
      exit sub
    end if
    i += 1
  nexteach
  if addIfMissing then elements.add(element)
end sub

sub TGui.show(state as integer = GuiState.active)
  dim as TGuiElement e
  foreach(TGuiElement, e, elements)
    e.show state
  nexteach
end sub
