type TWordPosition
  as integer x, w
  as integer size, special   ' not implemented yet
  as string text
  declare operator cast as string
end type

operator TWordPosition.cast as string
  return mki(x) & mki(w) & mki(len(text)) & text
end operator
operator =(w1 as TWordPosition, w2 as TWordPosition) as integer
  return str(w1) = str(w2)
end operator
mdListDeclare(TWordPosition)

type TTextLine
  as integer size, w
  as FT_Face font = font
  as mdList(TWordPosition) words
  declare operator cast as string
end type

operator TTextLine.cast as string
  return ""
end operator
operator =(w1 as TTextLine, w2 as TTextLine) as integer
  return str(w1) = str(w2)
end operator
mdListDeclare(TTextLine)

type TTextBlock
  as mdList(TTextLine) lines
  declare sub add(w as integer, t as string, font as FT_Face, size as integer = FONTSIZE.NORMAL)
  declare function show as FB.Image ptr
end type

sub TTextBlock.add(w as integer, t as string, font as FT_Face, size as integer = FONTSIZE.NORMAL)
  dim as integer position = 1, search, spaceWidth = FTsize(" ", font, size).x
  dim as integer wordWidth, lineWidth = 0
  dim as string currentWord
  dim as TTextLine currentLine
  dim as TWordPosition tempWord
  dim as single wordSpace
  currentLine.font = font
  currentLine.size = size
  currentLine.w    = w

  do
    search = instr(position+1, t, any chr(9, 32))
    if search then
      currentWord = mid(t, position, search-position)
      do
        search += 1
      loop until search >= len(t) orelse t[search-1] <> 9 and t[search-1] <> 32
      position = search
    else
      currentWord = mid(t, position)
      position = len(t) + 1
    end if
    wordWidth = FTsize(currentWord, font, size).x
    if lineWidth + wordWidth > w then
      wordSpace = (w-lineWidth-spaceWidth) / (currentLine.words.size-1)
      tempWord.text = ""
      currentLine.words.add(tempword)
      for i as integer = 1 to currentLine.words.size-1
        tempWord = currentLine.words.get(i)
        tempWord.x += i*wordSpace
        currentLine.words.set(i, tempword)
      next
      lines.add currentLine
      currentLine.words.clear
      tempWord.text = currentWord
      tempWord.x = 0
      currentLine.words.add tempWord
      lineWidth = wordWidth + spaceWidth
    else
      tempWord.text = currentWord
      tempWord.x = lineWidth
      currentLine.words.add tempWord
      lineWidth += wordWidth + spaceWidth
    end if
  loop while position < len(t)
  lines.add currentLine
end sub

function TTextBlock.show as FB.Image ptr
  dim as TTextLine currentLine
  dim as TWordPosition currentWord
  dim as integer spaceHeight, blockWidth = 0, blockHeight = 0, posY = 0
  for i as integer = 0 to lines.size-1
    currentLine = lines.get(i)
    blockHeight += FTsize(" ", currentLine.font, currentLine.size).y
    if blockWidth < currentLine.w then blockWidth = currentLine.w
  next
  dim as FB.Image ptr ret = imagecreate(blockWidth, blockHeight, rgb(255, 255, 255))
  for i as integer = 0 to lines.size-1
    currentLine = lines.get(i)
    spaceHeight = FTsize(" ", currentLine.font, currentLine.size).y
    for k as integer = 0 to currentLine.words.size-1
      currentWord = currentLine.words.get(k)
      printFT currentWord.x, posY, currentWord.text, currentLine.font, currentLine.size,, ret
    next
    posY += spaceHeight
  next
  return ret
end function
