' module: play "Hangman"

#include once "include/ploadlib.bi"

type THangmanDashData
  as integer y, w, h, dist, letterHeight, fSize
end type

namespace HANGMAN
  declare function callback(exercise as TExercise) as long
  declare function settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  declare sub showDash(wWordPtr as wstring ptr, position as integer, length as integer, _
                       dData as THangmanDashData, show as integer = -1)
end namespace

#ifndef HANGMAN_KEYLENGTH
  const HANGMAN_KEYLENGTH = 9
#endif

globalExercises.add(TExercise("hangman", localize("hangmanTitle"), "full", 0, _
                   @HANGMAN.callback, 0, @HANGMAN.settings, localize("hangmanHeadline"), "games"))

function HANGMAN.callback(exercise as TExercise) as long
  dim as FB.Image ptr img(100)
  dim as integer countImg
  dim as string wUtf8, path = MODULE_PATH & "img/"
  if fileExists(path & "hangman0.png") = 0 then path = exepath & "/img/hangman/"
  for countImg = -1 to 99
    if fileExists(path & "hangman" & countImg+1 & ".png") = 0 then exit for
    img(countImg+1) = pload(path & "hangman" & countImg+1 & ".png")
    if img(countImg+1) = 0 then exit for
  next
  if countImg < 1 then
    if img(0) then imagedestroy(img(0))
    message(localize("hangmanNoImg", path & "hangmanX.png"))
    return(RETVAR.dataError)
  end if
  dim as integer imgPosX = SCREEN_X - img(0)->width - 40, imgPosY = 100
  dim as integer LBxHelp = 50, LBxSetting = 90, LBxExit = 640, LBxSolve = 150, LBxNew = 300
  dim as integer LBy = 570, LBw = 120, LBh = 30
  dim as integer instrX = 50, instrY = 100, infX = 50, infY = 130, inf2X = 50, inf2Y = 170
  dim as integer specialKeysX = 50, specialKeysY = 250
  dim as integer dashY = 480, dashW = 15, dashH = 2, dashDist = 5, fsize = FONTSIZE.NORMAL
  dim as integer length, letterHeight = FTSize(" ", font, fsize).y
  dim as integer countHits, countGallows, tempInt, emptyDashes, mx, my, mb
  dim as string category, key, uKey, usedLetters
  dim as TWord word
  dim as wstring ptr wWordPtr
  dim as wstring*2 curLetter
  dim as THangmanDashData dashData = type<THangmanDashData>(dashY, dashW, dashH, dashDist, letterHeight, fSize)
  
  enum
    posExit = 1, posNew, posSolve, posHelp, posSetting
    posKeys = 100
  end enum

  ' prepare GUI
  dim as TGui gui
  dim as string keys()
  wordToArray EXTEND_UPPER, keys()
  guiAddKeys gui, posKeys, keys(), 50, 220, 30, 30, 10, 10, HANGMAN_KEYLENGTH
  dim as TGuiElement keyButton(ubound(keys))
  for i as integer = 0 to ubound(keys)
    keyButton(i) = gui.getElement(posKeys+i)
  next

  dim as TGuiEvent event
  gui.addElement _
      TGuiElement(GuiElementType.button, posHelp, _
                  exepath & "/img/help.bmp", LBxHelp, LBy-5, 40, 40, chr(255, 59),, "", GuiState.noBorder)
  gui.addElement _
      TGuiElement(GuiElementType.button, posSetting, _
                  exepath & "/img/settings.bmp", LBxSetting, LBy-5, 40, 40, chr(9),, "", GuiState.noBorder)
  gui.addElement _
      TGuiElement(GuiElementType.button, posSolve, LBxSolve, LBy, LBw, LBh, localize("solution"))
  gui.addElement _
      TGuiElement(GuiElementType.button, posNew, LBxNew, LBy, LBw, LBh, localize("newPuzzle"))
  gui.addElement _
      TGuiElement(GuiElementType.button, posExit, LBxExit, LBy, LBw, LBh, localize("back"), chr(27))

  do
    countGallows = 0
    emptyDashes  = 0
    usedLetters  = ""
    if exercise.currentList.size then
      word = randomWord(exercise.currentList)
    else
      word = randomWord(exercise.fullList)
    end if
    wUtf8 = modifyTxt(word.word, MODIFY_TOUPPER or MODIFY_NOACUTE)
    length = wordLength(wUtf8)
    wWordPtr = callocate((length+1)*sizeof(wstring))
    UTFToWChar(UTF_ENCOD.UTF_ENCOD_UTF8, strptr(wUtf8), wWordPtr, @length)
    if word.tag.size > 0 then
      category = word.tag.get(int(rnd*word.tag.size))
    else
      category = localize("group-" & word.groupname)
    end if

    for i as integer = 0 to ubound(keys)
      keyButton(i).state = GuiState.active
      gui.setElement keyButton(i)
    next

    screenlock
    cls
    centerFT SCREEN_X\2, 50, exercise.headline, font, 20
    put (imgPosX, imgPosY), img(0), pset

    printFT instrX, instrY, localize("hangmanInstruct"), font
    printFT infX, infY, localize("hangmanInfo", category), font
    printFT inf2X, inf2Y, localize("hangmanInfo2", category), font, FONTSIZE.TINY
    for i as integer = 0 to length-1
      tempInt = instr(EXTEND_UPPER, mid(*wWordPtr, i+1, 1))
      if tempInt then emptyDashes += 1
      showDash wWordPtr, i, length, dashData, tempInt = 0
    next
    gui.show
    screenunlock
    
    ' check GUI / enter letters
    do
      getmouse mx, my,, mb
      sleep 1
    loop until mb = 0
    do
      key = inkey
      event = gui.getEvent(key)
      select case event.id
        case posHelp
          tempInt = showHelpScreen(MODULE_PATH & "doc.htm", "hangman")
          if tempInt = RETVAR.windowClosed then key = chr(255, 107)
        case posExit
          exit do, do
        case posNew
          exit do
        case posSolve
          for i as integer = 0 to length-1
            showDash wWordPtr, i, length, dashData
          next
          emptyDashes = 0
          message localize("hangmanSolution", word.word, word.translate.get(0)),, 1
        case posKeys to posKeys + ubound(keys)
          key = keys(event.id - posKeys)
      end select
      uKey = ucase(key)
      if emptyDashes andalso len(key) then
        if instr(EXTEND_UPPER, uKey) andalso instr(usedLetters, uKey) = 0 then
          countHits = 0
          for i as integer = 0 to length-1
            curLetter = wWordPtr[i]
            if curLetter = uKey then
              showDash wWordPtr, i, length, dashData
              countHits += 1
              emptyDashes -= 1
            end if
          next
          if countHits = 0 then
            countGallows += 1
            put (imgPosX, imgPosY), img(countGallows), pset
          end if
          usedLetters &= uKey
          for i as integer = 0 to ubound(keys)
            if keys(i) = uKey then
              keyButton(i).state = GuiState.activated
              gui.setElement keyButton(i)
              keyButton(i).show
              exit for
            end if
          next
        end if
      end if
      sleep 1
    loop until key = chr(255, 107)
    deallocate wWordPtr
  loop until key = chr(255, 107)
  
  for i as integer = 0 to countImg
    imagedestroy img(i)
  next
  select case key
    case chr(255, 107) : return RETVAR.windowClosed
    case else          : return RETVAR.ok
  end select
end function

function HANGMAN.settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  return chooseWords(exercise, exercise.settingString, optional or SETTING_OPTIONS.SHOWTAGS)
end function

sub HANGMAN.showDash(wWordPtr as wstring ptr, position as integer, length as integer, _
                     dData as THangmanDashData, show as integer = -1)
  dim as wstring*2 curLetter = wWordPtr[position]
  if curLetter <> " " then line((SCREEN_X-(length-2*position)*(dData.w+dData.dist))\2, dData.y)-step(dData.w, dData.h), 0, bf
  if show then centerFT (SCREEN_X-(length-2*position)*(dData.w+dData.dist) + dData.w)\2 - 2, _
                         dData.y - dData.letterHeight\2, curLetter, font, dData.fSize
end sub
