' basic module for language training Spanish-German

const SCREEN_X = 800, SCREEN_Y = 640, B_COUNT = 6
const moduleVersionMajor = 2, moduleVersionMinor = 3, moduleVersionPatch = 170721
const LANGTITLE1 = "Spanisch",   LANGTITLE2 = "Deutsch"
const LANGSHORT1 = "es",         LANGSHORT2 = "de"

dim shared as string LANGUAGE, MODULE_PATH, WORDS_VERSION
LANGUAGE = mid(__PATH__, instrrev(__PATH__, any "/\")+1)
MODULE_PATH = exepath & "/modules/" & LANGUAGE & "/"

' options to modify text to a different form
' BE SURE NOT TO USE CONFLICTING OPTIONS!
const MODIFY_TOLOWER       =   1, _ ' change to lower case if possible
      MODIFY_TOUPPER       =   2, _ ' change to upper case if possible
      MODIFY_NOPARANTHESES =   4, _ ' omit parantheses ( and ) 
      MODIFY_NOSPECIAL     =   8, _ ' omit everything exept a-z, A-Z, 0-9 and language related keys
                                  _ ' (e. g. umlauts in German)
      MODIFY_NOACUTE       =  16, _ ' change é to e and so on (umlauts remain unaffected)
      MODIFY_NOUMLAUT      =  32, _ ' change ß to ss (or SS if MODIFY_TOUPPER), Ä to Ae (or AE or ae) and so on
      MODIFY_NOTILDE       =  64, _ ' change ñ/Ñ to n/N
      MODIFY_NONUMBERS     = 128, _ ' omit 0-9
      MODIFY_SORTABLE      = 256, _ ' various changes like lower case, ä->ae, é->e, ñ->n|
                                  _ ' (to be sorted between n and o) ...
      MODIFY_ALPHANUM      = 512    ' everthing exept a-z, A-Z and 0-9 is changed to an underline character

' inflection forms
namespace INFLECT
  const PERS1 = 1, PERS2 = 2, PERS3 = 3
  const SINGULAR = 0, PLURAL = 3
  const PRES = 0, PRETIND = 8, PRETIMP = 16, PERF = 24, PQPERF = 32, PERFANT = 40
  const FUTSIMP = 48, FUTCOMP = 56, IMPERAT = 64, IMPERATNO = 72, COND = 80, CONDPERF = 88
  const PARTIC = 96, GERUND = 104, SUBJ = 128
  const MASC = 0, FEM = 1, GENUS = 1, SING = 0, PL = 2, NUM = 2, ADVERB = 4
  const FULL = 1, MINIMUM = 2
end namespace

const COLOR_GRAY = rgb(127, 127, 127)

#include once "include/font.bi"
dim shared font as FT_Face

#include once "include/globalFunctions.bi"
#include once "include/TExercise.bi"
#include once "include/TInput.bi"
#include once "include/showWords.bi"
#include once "include/help.bi"
#include once "include/levenshtein.bi"
#include once "include/multiput.bi"
dim shared as TExercise globalExercArray(), globalOptArray()

declare function addAcute(char as string) as string
declare function conjugSpanish(verb as TWord, inflection as integer) as TWord
declare function createFlexion(w as TWord, byref flex as integer = 0) as TWord
declare function declinSpanish(noun as TWord, inflection as integer) as string
declare function firstUpper(txt as string) as string
declare function flexionCount(w as TWord, opt as integer = INFLECT.FULL) as integer
declare function getArticle(w as TWord, both as integer = 0) as string
declare sub init(mainModule as TModule ptr)
declare function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
declare function number(first as integer = 0, last as integer = 0) as TWord
declare function pluralSpanish(sing as string) as string
declare sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
declare sub speakThread(txt as string, no as integer = 1)
declare sub wordInfo(w as TWord)

declare function conjugGerman(verb as TWord, flex as integer, no as integer = 0) as string
declare function declinGerman(noun as TWord, flex as integer, no as integer = 0) as string
#include once "specific.bi"
#include once "include/common/lang-basic.bi"
#include once "addGerman.bi"

#include once "include/common/editWords.bi"
#include once "include/common/learning.bi"
#include once "include/common/wordtest.bi"
#include once "include/common/diagram.bi"
#include once "include/common/dictionary.bi"
#include once "export.bi"
#include once "grammarScheme.bi"
'#include once "include/common/similar.bi"
#include once "conjugation.bi"
#include once "incgui.bi"

' games
#include once "include/common/pairs.bi"
#include once "include/common/crossword.bi"
#include once "include/common/hangman.bi"
globalExercises.add(TExercise("games", localize("gamesTitle"), "full", MENUTYPE.exercTwoRowsExt, _
                    @createCategory, 0, 0, localize("gamesHeadline")))

' grammar
#include once "grammar.bi"
globalExercises.add(TExercise("grammar", localize("grammarTitle"), "full", MENUTYPE.exercTwoRowsExt, _
                    @createCategory, @GRAMMAR.info, 0, localize("grammarHeadline")))

function addAcute(char as string) as string
  ' "a" -> "á"; "e" -> "é" etc.
  select case char
    case "a" : return "á"
    case "e" : return "é"
    case "i" : return "i"
    case "o" : return "ó"
    case "u" : return "ú"
  end select
  return char
end function

function conjugSpanish(verb as TWord, inflection as integer) as TWord
  dim as TWord ret
  if len(verb.word) < 2 then return ret    ' error

  dim as integer declgroup, mark = 0, person = (inflection and 7)-1, tense, group, specialEnding
  dim as string retWord, tempString, root, regularEnding
  dim as string infl = ((person mod INFLECT.PLURAL) + 1) _
                       & iif(person\INFLECT.PLURAL, "p", "s"), reflex = ""
  dim as string ending, regularRoot = verb.map.get("root")
  select case person+1
    case INFLECT.PERS1 + INFLECT.SINGULAR : person = 0
    case INFLECT.PERS2 + INFLECT.SINGULAR : person = 1
    case INFLECT.PERS3 + INFLECT.SINGULAR : person = 2
    case INFLECT.PERS1 + INFLECT.PLURAL   : person = 3
    case INFLECT.PERS2 + INFLECT.PLURAL   : person = 4
    case INFLECT.PERS3 + INFLECT.PLURAL   : person = 5
  end select
  if len(regularRoot) = 0 then regularRoot = verb.word
  if right(regularRoot, 2) = "se" then
    if len(regularRoot) < 4 then return ret
    select case as const person
      case 0 : reflex = "me "
      case 1 : reflex = "te "
      case 2 : reflex = "se "
      case 3 : reflex = "nos "
      case 4 : reflex = "os "
      case 5 : reflex = "se "
    end select
    declgroup = regularRoot[len(regularRoot)-4]
    regularRoot = left(regularRoot, len(regularRoot)-4)
  elseif right(regularRoot, 3) = "ír" then
    if len(regularRoot) < 3 then return ret
    declGroup = asc("i")
    regularRoot = left(regularRoot, len(regularRoot)-3)
    specialEnding = 1
  else
    if len(regularRoot) < 2 then return ret
    declGroup = regularRoot[len(regularRoot)-2]
    regularRoot = left(regularRoot, len(regularRoot)-2)
  end if
  
  ' composed / participle
  select case inflection and not 7 and not INFLECT.SUBJ
    case INFLECT.PERF, INFLECT.PQPERF, INFLECT.PERFANT, INFLECT.FUTCOMP, INFLECT.CONDPERF, INFLECT.PARTIC
      dim as string participle = verb.map.get("wparticiple"), retword
      if len(participle) then
        mark = -1
      else
        participle = verb.map.get("root")
        if len(participle) then
          mark = -1
          participle = left(participle, len(participle)-1+2*(right(participle, 2) = "se")) & "do"
        else
          participle = left(verb.word, len(verb.word)-1+2*(right(verb.word, 2) = "se")) & "do"
        end if
        if right(participle, 3) = "edo" then participle = left(participle, len(participle)-3) & "ido"
        if right(participle, 3) = "ido" andalso instr("aeiou", mid(participle, len(participle)-3, 1)) then
          if right(participle, 5) <> "guido" and right(participle, 5) <> "quido" then
            participle = left(participle, len(participle)-3) & "ído"
          end if
        end if
      end if
      select case inflection and not 7
        case INFLECT.PARTIC
          retword = participle
        case INFLECT.PERF
          dim as string pre(...) = { "he", "has", "ha", "hemos", "habéis", "han" }
          retword = pre(person) & " " & participle
        case INFLECT.PQPERF
          dim as string pre(...) = { "había", "habías", "había", "habíamos", "habíais", "habían" }
          retword = pre(person) & " " & participle
        case INFLECT.PERFANT
          dim as string pre(...) = { "hube", "hubiste", "hubo", "hubimos", "hubisteis", "hubieron" }
          retword = pre(person) & " " & participle
        case INFLECT.FUTCOMP
          dim as string pre(...) = { "habré", "habrás", "habrá", "habremos", "habréis", "habran" }
          retword = pre(person) & " " & participle
        case INFLECT.CONDPERF
          dim as string pre(...) = { "habría", "habrías", "habría", "habríamos", "habríais", "habrían" }
          retword = pre(person) & " " & participle
        case INFLECT.PERF or INFLECT.SUBJ
          dim as string pre(...) = { "haya", "hayas", "haya", "hayamos", "hayáis", "hayan" }
          retword = pre(person) & " " & participle
        case INFLECT.PQPERF or INFLECT.SUBJ
          dim as string pre(...) = { "hubiera", "hubieras", "hubiera", "hubiéramos", "hubierais", "hubieran" }
          retword = pre(person) & " " & participle
        case INFLECT.FUTCOMP or INFLECT.SUBJ
          dim as string pre(...) = { "hubiere", "hubieres", "hubiere", "hubiéremos", "hubiereis", "hubieren" }
          retword = pre(person) & " " & participle
      end select
      ret.word = reflex & retWord
      ret.mark = mark
      return ret
  end select

  ' declined
  dim as string endings(2, 10, 5) = { _
      { { "o", "as", "a", "amos", "áis", "an"},                 { "aba", "abas", "aba", "ábamos", "abais", "aban" }, _
        { "é", "aste", "ó", "amos", "asteis", "aron"},          { "aré", "arás", "ará", "aremos", "aréis", "arán" }, _
        { "e", "es", "e", "emos", "éis", "en"},                 { "ara", "aras", "ara", "áramos", "arais", "aran" }, _
        { "ase", "ases", "ase", "ásemos", "aseis", "asen"},     { "are", "ares", "are", "áremos", "areis", "aren" }, _
        { "aría","arías", "aría", "aríamos", "aríais", "arían" }, _
        { "", "a", "e", "emos", "ad", "en" },                   { "ar", "ando", "ado", "ada", "ados", "adas" } }, _
      { { "o", "es", "e", "emos", "éis", "en" },                { "ía", "ías", "ía", "íamos", "íais", "ían" }, _
        { "í", "iste", "ió", "imos", "isteis", "ieron" },       { "eré", "erás", "erá", "eremos", "eréis", "erán" }, _
        { "a", "as", "a", "amos", "áis", "an"},                 { "era", "eras", "era", "éramos", "erais", "eran" }, _
        { "ese", "eses", "ese", "ésemos", "eseis", "esen" },    { "ere", "eres", "ere", "éremos", "ereis", "eren" }, _
        { "ería","erías", "ería","eríamos", "eríais", "erían" }, _
        { "", "e", "a", "amos", "ed", "an" },                   { "er", "iendo", "ido", "ida", "idos", "idas" } }, _
      { { "o", "es", "e", "imos", "ís", "en" },                 { "ía", "ías", "ía", "íamos", "íais", "ían" }, _
        { "í", "iste", "ió", "imos", "isteis", "ieron" },       { "iré", "irás", "irá", "iremos", "iréis", "irán" }, _
        { "a", "as", "a", "amos", "áis", "an"},                 { "era", "eras", "era", "éramos", "erais", "eran" }, _
        { "ese", "eses", "ese", "ésemos", "eseis", "esen" }, { "ere", "eres", "ere", "éremos", "ereis", "eren" }, _
        { "iría", "irías", "iría","iríamos", "iríais", "irían" }, _
        { "", "e", "a", "amos", "id", "an" },                   { "ir", "iendo", "ido", "ida", "idos", "idas" } } }
  ' looking for irregular root
  select case inflection and (not 7)
    case INFLECT.PRES
      tense = 0
      infl &= "pres"
      root = verb.map.get("rootpres")
      if len(root) then mark = -1
      if len(root) = 0 andalso person <> 3 and person <> 4 then
        root = verb.map.get("rootdiph")
        if len(root) then
          mark = -1
          regularRoot = root
          root = ""
          declgroup = asc(regularRoot, len(regularRoot))
          regularRoot = left(regularRoot, len(regularRoot)-1)
        end if
      end if
    case INFLECT.PRETIMP
      tense = 1
      infl &= "pretimp"
      root = verb.map.get("rootpretimp")
      if len(root) then mark = -1
    case INFLECT.PRETIND
      tense = 2
      infl &= "pretind"
      root = verb.map.get("rootpretind")
      if len(root) then mark = -1
    case INFLECT.FUTSIMP
      tense = 3
      infl &= "fut"
      root = verb.map.get("rootfut")
      if len(root) then mark = -1
    case INFLECT.PRES or INFLECT.SUBJ
      tense = 4
      infl &= "pressubj"
      root = verb.map.get("rootpressubj")
      if len(root) then
        mark = -1
      else
        root = verb.map.get("w1spres")
        if len(root) then
          root = left(root, len(root)-1) & mid(root, len(root)-1, 1)
          mark = -1
        end if
      end if
      if len(root) = 0 andalso person <> 3 and person <> 4 then
        root = verb.map.get("rootdiph")
        if len(root) then
          mark = -1
          regularRoot = root
          root = ""
          declgroup = asc(regularRoot, len(regularRoot))
          regularRoot = left(regularRoot, len(regularRoot)-1)
        end if
      end if
    case INFLECT.PRETIMP or INFLECT.SUBJ
      tense = 5
      infl &= "pretsubj"
      dim as TWord tempWord = conjugSpanish(verb, INFLECT.PRETIND or (INFLECT.PERS3 + INFLECT.PLURAl))
      root = left(tempWord.word, len(tempWord.word)-3)
      dim as integer search = instr(root, " ")
      if search then root = mid(root, search + 1)
      if tempWord.mark then mark = -1
    case INFLECT.PRETIND or INFLECT.SUBJ   ' alternative form for pret. imp. subj.
      tense = 6
      infl &= "pretsubj2"
      dim as TWord tempWord = conjugSpanish(verb, INFLECT.PRETIND or (INFLECT.PERS3 + INFLECT.PLURAl))
      root = left(tempWord.word, len(tempWord.word)-3)
      dim as integer search = instr(root, " ")
      if search then root = mid(root, search + 1)
      if tempWord.mark then mark = -1
    case INFLECT.FUTSIMP or INFLECT.SUBJ
      tense = 7
      infl &= "futsubj"
      dim as TWord tempWord = conjugSpanish(verb, INFLECT.PRETIND or (INFLECT.PERS3 + INFLECT.PLURAl))
      root = left(tempWord.word, len(tempWord.word)-3)
      dim as integer search = instr(root, " ")
      if search then root = mid(root, search + 1)
      if tempWord.mark then mark = -1
    case INFLECT.COND
      tense = 8
      infl &= "cond"
      root = verb.map.get("rootcond")
      if len(root) then mark = -1
    case INFLECT.IMPERAT
      dim as TWord tempWord
      if person = 1 then
        root = verb.map.get("wimperat")
        if len(root) then
          verb.word = root
        else
          tempWord = conjugSpanish(verb, (INFLECT.PERS3+INFLECT.SINGULAR) or INFLECT.PRES)
          if left(tempWord.word, 3) = "se " then tempWord.word = mid(tempWord.word, 4)
        end if
      elseif person = 4 then
        root = verb.map.get("wimperatp")
        if len(root) then
          tempWord.word = root
          tempWord.mark = -1
          return tempWord
        else
          if specialEnding = 1 then
            verb.word = regularRoot & "íd"
          else
            tempWord = TWord(regularRoot & chr(declGroup) & "d", declinGerman(verb, inflection, 0), "verb")
            verb.word = regularRoot & chr(declGroup) & "d"
          end if
        end if
      else
        tempWord = conjugSpanish(verb, (inflection and 7) or INFLECT.PRES or INFLECT.SUBJ)
        if left(tempWord.word, 3) = "se " then
          tempWord.word = mid(tempWord.word, 4)
        elseif left(tempWord.word, 4) = "nos " then
          tempWord.word = mid(tempWord.word, 5)
        end if
      end if      
      if len(reflex) then
        if person = 4 then
          tempWord.word = left(tempWord.word, len(tempWord.word)-1) & reflex
        else
          dim as integer vocal = instrrev(tempWord.word, any "aeiou")
          vocal = instrrev(tempWord.word, any "aeiou", vocal-1)
          if vocal then
            select case mid(tempWord.word, vocal, 1)
              case "a" : tempWord.word = left(tempWord.word, vocal-1) & "á" & mid(tempWord.word, vocal+1)
              case "e" : tempWord.word = left(tempWord.word, vocal-1) & "é" & mid(tempWord.word, vocal+1)
              case "i" : tempWord.word = left(tempWord.word, vocal-1) & "í" & mid(tempWord.word, vocal+1)
              case "o" : tempWord.word = left(tempWord.word, vocal-1) & "ó" & mid(tempWord.word, vocal+1)
              case "u" : tempWord.word = left(tempWord.word, vocal-1) & "ú" & mid(tempWord.word, vocal+1)
            end select
          end if
          if person = 3 then
            tempWord.word = left(tempWord.word, len(tempWord.word)-1) & reflex
          else
            tempWord.word &= reflex
          end if
        end if
      end if
      return tempword
    case INFLECT.GERUND
      tense = 10
      person = 1
      infl = verb.map.get("wgerund")
      if len(infl) then
        ret.word = infl
        ret.mark = -1
        return ret
      end if
  end select

  if len(root) then
    declgroup = root[len(root)-1]
    regularRoot = left(root, len(root)-1)
  end if
  group = iif(declgroup = 97, 0, iif(declgroup = 101, 1, 2))
  ending = endings(group, tense, person)

  ' add accent or -y- if necessary
  if specialEnding <> 1 andalso instr("aeiéí", right(regularRoot, 1)) then
    if ending = "iste" then
      ending = "íste"
    elseif ending = "imos" then
      ending = "ímos"
    elseif ending = "isteis" then
      ending = "ísteis"
    elseif left(ending, 1) = "i" then
      ending = "y" & mid(ending, 2)
    end if
  end if
  ' delete redundant accents
  if instr(regularRoot, any "aeiou") = 0 then
    select case ending
      case "áis" : ending = "ais"
      case "éis" : ending = "eis"
      case "ís"  : ending = "is"
      case "é"   : ending = "e"
      case "í"   : ending = "i"
      case "ó"   : ending = "o"
      case "ió"  : ending = "io"
    end select
  end if

  ' regular changes
  if person <> 3 and person <> 4 then
    select case inflection and (not 7)
      case INFLECT.PRES, INFLECT.PRES or INFLECT.SUBJ, INFLECT.IMPERAT
        if right(regularRoot, 1) = "u" andalso right(regularRoot, 2) <> "gu" then
          regularRoot = left(regularRoot, len(regularRoot)-1) & "ú"
        end if
    end select
  end if
  if group = 0 then
    if right(regularRoot, 1) = "c" andalso left(ending, 1) = "e" or left(ending, 2) = "é" then
      regularRoot = left(regularRoot, len(regularRoot)-1) & "qu"
    end if
    if right(regularRoot, 1) = "z" andalso _
          left(ending, 1) = "e" or left(ending, 2) = "é" or left(ending, 1) = "i" or left(ending, 2) = "í" then
      regularRoot = left(regularRoot, len(regularRoot)-1) & "c"
    end if
    if right(regularRoot, 1) = "g" andalso left(ending, 1) = "e" or left(ending, 2) = "é" then
      regularRoot = left(regularRoot, len(regularRoot)-1) & "gu"
    end if
  else
    if right(regularRoot, 1) = "c" then
      if left(ending, 1) = "o" or left(ending, 1) = "a" or left(ending, 2) = "á" then
        regularRoot = left(regularRoot, len(regularRoot)-1) & "zc"
      end if
    end if
    if right(regularRoot, 1) = "g" then
      if left(ending, 1) = "o" or left(ending, 1) = "a" or left(ending, 2) = "á" then
        regularRoot = left(regularRoot, len(regularRoot)-1) & "j"
      end if
    end if
  end if
  if right(regularRoot, 2) = "gu" then
    if left(ending, 1) = "a" or left(ending, 1) = "o" or left(ending, 2) = "á" or left(ending, 2) = "ó" then
      regularRoot = left(regularRoot, len(regularRoot)-1)
    end if
  end if

  ' looking for irregular form
  retWord  = verb.map.get("w" & infl)
  if len(retWord) then
    mark = -1
  else
    if len(root) then
      root = left(root, len(root)-1)
    else
      root = regularRoot
    end if
    if len(root) = 0 then root = regularRoot
    retWord = root & ending
  end if

  ' different forms
  retWord = reflex & retWord
  ret.word = retWord

  ret.mark = mark
  return ret
end function

function createFlexion(w as TWord, byref flex as integer = 0) as TWord
  ' create random flexion of the word (e. g. for crossword.bi)
  dim ret as TWord = w, translation as string, tempString as string
  dim as integer transRand = int(rnd*w.translate.size)
  ret.word = w.word
  select case w.groupname
    case "verb"
      flex = int(rnd*6) + 1
      dim as TWord tempWord = conjugSpanish(w, flex)
      ret.word = tempWord.word
    case "noun", "adjective"
      dim as string genus = w.map.get("genus")
      dim as string fem, num   = w.map.get("num")
      if genus = "m" then fem = w.map.get("fem") else fem = w.word
      select case num
        case "sg" : flex = INFLECT.SING
        case "pl" : flex = INFLECT.PL
        case else : flex = iif(rnd < .5, INFLECT.SING, INFLECT.PL)
      end select
      if genus = "m" then
        if len(fem) then
          flex or= iif(rnd < .5, INFLECT.MASC, INFLECT.FEM)
        else
          flex or= INFLECT.MASC
        end if
      elseif genus = "f" then
        flex or= INFLECT.FEM
      else
        flex or= iif(rnd < .5, INFLECT.MASC, INFLECT.FEM)
      end if
      ret.word =  declinSpanish(w, flex)
  end select
  return ret
end function

function declinSpanish(noun as TWord, inflection as integer) as string
  dim as string num, f, pl, genus = noun.map.get("genus")
  if (inflection and INFLECT.GENUS) = INFLECT.MASC then
    if genus = "f" then return ""
    num = noun.map.get("num")
    if (inflection and INFLECT.NUM) = INFLECT.SING then
      return iif(num = "pl", "", noun.word)
    else
      if num = "sg" then return ""
      if num = "pl" then return noun.word
      pl = noun.map.get("pl")
      return iif(len(pl), pl, pluralSpanish(noun.word))
    end if
  else
    if genus = "m" then
      f = noun.map.get("f")
      if len(f) = 0 then return ""
      if (inflection and INFLECT.NUM) = INFLECT.SING then
        return iif(num = "pl", "", f)
      else
        if num = "sg" then return ""
        if num = "pl" then return f
        pl = noun.map.get("pl-f")
        return iif(len(pl), pl, pluralSpanish(f))
      end if
    else
      num = noun.map.get("num")
      if (inflection and INFLECT.NUM) = INFLECT.SING then
        return iif(num = "pl", "", noun.word)
      else
        if num = "sg" then return ""
        if num = "pl" then return noun.word
        pl = noun.map.get("pl")
        return iif(len(pl), pl, pluralSpanish(noun.word))
      end if
    end if
  end if
end function

function firstUpper(txt as string) as string
  if asc(txt) < 128 then
    return ucase(left(txt, 1)) & mid(txt, 2)
  else
    select case left(txt, 2)
      case "ä", "ö", "ü", "á", "é", "í", "ó", "ú", "ñ"
        return chr(txt[0]) & chr(txt[1]-32) & mid(txt, 3)
    end select
  end if
  return txt
end function

function flexionCount(w as TWord, opt as integer = INFLECT.FULL) as integer
  select case w.groupname
    case "verb"
      select case opt
        case INFLECT.MINIMUM : return 6
        case else            : return 255
      end select
    case "noun"      : return 3
    case "adjective" : return 4
    case else        : return -1
  end select
end function

function getArticle(w as TWord, both as integer = 0) as string
  ' returns the article(s) of a noun, followed by a space character
  ' If there is no article (because w is not a noun), the function returns an empty string
  ' both <> 0: if there exist (different) forms for m and f, return el/la, or les/las respectively
  if w.groupname <> "noun" then return ""
  dim as string genus = w.map.get("genus"), article, num
  if len(genus) then article = w.map.get("article")
  if len(article) then return article & " "   ' special article (e. g. 'el aula')
  num = w.map.get("num")
  select case genus
    case "m"
      if both andalso len(w.map.get("f")) then
        article &= iif(num = "pl", "los/las ", "el/la ")
      else
        article = iif(num = "pl", "los ", "el ")
      end if
    case "f"
      article = iif(num = "pl", "las ", "la ")
    case else
      if both then
        article = iif(num = "pl", "los/las ", "el/la ")
      else
        article = ""
      end if
  end select
  return article
end function

sub init(mainModule as TModule ptr) export
  initModule(mainModule)
  dim as TList fullList
  fullList.listname = "complete"
  fullList.add , 0, &hffffff
  fullList.save
  fullList.listname = "full"
  fullList.del 0, &hffffff, "group", "cardinal"
  fullList.save
  initDefaultLists(fullList)
  mainWindow
  exitModule
end sub

function modifyTxt(txt as string, opt as integer = MODIFY_TOLOWER or MODIFY_NOPARANTHESES) as string
  ' modify input (see modify options at the top of the file)
  ' Using conflicting options may produce unexpected results
  dim as string ret
  dim as integer i = 0
  do until i >= len(txt)
    select case as const txt[i]
      case 0 to 31                                    ' ignore control chars
      case 65 to 90                                   ' upper case
        ret &= chr(txt[i] + iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), 32, 0))
      case 97 to 122                                  ' lower case
        ret &= chr(txt[i] - iif(opt and MODIFY_TOUPPER, 32, 0))
      case 40, 41                                     ' parentheses
        if opt and MODIFY_ALPHANUM then
          ret &= "_"
        else
          if opt and (MODIFY_NOPARANTHESES or MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
        end if
      case 48 to 57                                   ' numbers
        if (opt and MODIFY_NONUMBERS) = 0 then ret &= chr(txt[i])
      case 32 to 39, 42 to 47, 58 to 64, 91 to 96, 123 to 127   ' special keys
        if opt and MODIFY_ALPHANUM then
          ret &= "_"
        else
          if (opt and MODIFY_NOSPECIAL) = 0 then ret &= chr(txt[i])
        end if
      case 195
        i += 1
        if i >= len(txt) then exit do
        if opt and MODIFY_ALPHANUM then ret &= "_" : exit select
        select case as const txt[i]
          case 132 ' Ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ae"
              elseif opt and MODIFY_TOUPPER then
                ret &= "AE"
              else
                ret &= "Ae"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ä", "Ä")
            end if
          case 164 ' ä
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "AE", "ae")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ä", "ä")
            end if
          case 150 ' Ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "oe"
              elseif opt and MODIFY_TOUPPER then
                ret &= "OE"
              else
                ret &= "Oe"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ö", "Ö")
            end if
          case 182 ' ö
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "OE", "oe")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ö", "ö")
            end if
          case 156 ' Ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              if opt and (MODIFY_SORTABLE or MODIFY_TOLOWER) then
                ret &= "ue"
              elseif opt and MODIFY_TOUPPER then
                ret &= "UE"
              else
                ret &= "Ue"
              end if
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ü", "Ü")
            end if
          case 188 ' ü
            if opt and MODIFY_SORTABLE or MODIFY_NOUMLAUT then
              ret &= iif(opt and MODIFY_TOUPPER, "UE", "ue")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ü", "ü")
            end if
          case 159 ' ß
            if opt and (MODIFY_SORTABLE or MODIFY_NOUMLAUT) then
              ret &= iif(opt and MODIFY_TOUPPER, "SS", "ss")
            else
              ret &= "ß"
            end if
          case 129 ' Á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "a", "A")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "á", "Á")
            end if
          case 161 ' á
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "A", "a")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Á", "á")
            end if
          case 137 ' É
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "e", "E")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "é", "É")
            end if
          case 169 ' é
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "E", "e")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "É", "é")
            end if
          case 141 ' Í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "i", "I")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "í", "Í")
            end if
          case 173 ' í
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "I", "i")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "´I", "´i")
            end if
          case 147 ' Ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "o", "O")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ó", "Ó")
            end if
          case 179 ' ó
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "O", "o")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ó", "ó")
            end if
          case 154 ' Ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and (MODIFY_TOLOWER or MODIFY_SORTABLE), "u", "U")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ú", "Ú")
            end if
          case 186 ' ú
            if opt and (MODIFY_SORTABLE or MODIFY_NOACUTE) then
              ret &= iif(opt and MODIFY_TOUPPER, "U", "u")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ú", "ú")
            end if
          case 145 ' Ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOLOWER, "n|", "N|")
            elseif opt and MODIFY_NOTILDE then
              ret &= iif(opt and MODIFY_TOLOWER, "n", "N")
            else
              ret &= iif(opt and MODIFY_TOLOWER, "ñ", "Ñ")
            end if
          case 177 ' ñ
            if opt and MODIFY_SORTABLE then
              ret &= iif(opt and MODIFY_TOUPPER, "N|", "n|")
            elseif opt and MODIFY_NOTILDE then
              ret &= iif(opt and MODIFY_TOUPPER, "N", "n")
            else
              ret &= iif(opt and MODIFY_TOUPPER, "Ñ", "ñ")
            end if
        end select
      case 194
        i += 1
        if i >= len(txt) then exit do
        if opt and MODIFY_ALPHANUM then ret &= "_" : exit select
        select case as const txt[i]
          case 161     ' ¡
            if (opt and (MODIFY_SORTABLE or MODIFY_NOSPECIAL)) = 0 then ret &= "¡"
          case 191     ' ¿
            if (opt and (MODIFY_SORTABLE or MODIFY_NOSPECIAL)) = 0 then ret &= "¿"
        end select
    end select
    i += 1
  loop
  return ret
end function

function number(first as integer = 0, last as integer = 0) as TWord
  ' create a random number from 0 to 999'999'999
  ' (greater numbers first will give back a wrong german translation)
  ' if "last" isn't greater than "first", give back the word for "first" (no random number)
  dim as integer symbol, no = iif(last > first, int(rnd*(last-first+1)) + first, first)
  dim as string retWord, retTrans
  dim as TWord ret, w, w1000, w100, w10
  static as TList list_numbers = TList("numbers")
  if list_numbers.list.isEmpty then list_numbers.add , 0, &hffffff, "group", "cardinal"

  ' numbers over 1 million
  if no > 999999 then
    w = searchWord(list_numbers, "1000000", "symbol")
    if no < 2000000 then
      retWord  = w.word
      retTrans = w.translate.get(0)
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no\1000000)
      retWord  = w.word & " " & retWord
      retTrans = w.translate.get(0) & " " & retTrans
    end if
    if no mod 1000000 > 0 then
      w = number(no mod 1000000)
      retWord  &= " " & w.word
      retTrans &= " " & w.translate.get(0)    
    end if
    ret = TWord(retWord, retTrans, "cardinal")
    ret.map.put("symbol", str(no))
    return ret
  end if

  ' numbers from 1'000 to 999'999
  if no > 999 then
    w = searchWord(list_numbers, "1000", "symbol")
    if no mod 1000 = 0 then
      retWord  = w.word
      retTrans = w.translate.get(0)
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no mod 1000)
      retWord  = replace(retWord,  "%", w.word)
      retTrans = replace(retTrans, "%", _
                 iif(no mod 1000 > 99 and no mod 1000 < 200, "ein", "") & w.translate.get(0))
    end if
    if no > 1999 then
      w = number(no\1000)
      retWord =  w.word & " " & retWord
      retTrans = w.translate.get(0) & retTrans
    end if
    ret = TWord(retWord, retTrans, "cardinal")
    ret.map.put("symbol", str(no))
    return ret
  end if

  ' numbers from 100 to 999
  if no > 99 then
    w = searchWord(list_numbers, no\100 & "00", "symbol")
    if no mod 100 = 0 then
      return w
    else
      retWord  = w.map.get("generic-word")
      retTrans = w.map.get("generic-trans")
      w = number(no mod 100)
      retWord  = replace(retWord,  "%", w.word)
      retTrans = replace(retTrans, "%", w.translate.get(0))
      ret = TWord(retWord, retTrans, "cardinal")
      ret.map.put("symbol", str(no))
      return ret
    end if
  end if

  ' numbers from 0 to 99
  w = searchWord(list_numbers, str(no), "symbol")
  if w.id then return w
  w = searchWord(list_numbers, no\10 & "0", "symbol")
  retWord  = w.map.get("generic-word")
  retTrans = w.map.get("generic-trans")
  w = number(no mod 10)
  retWord  = replace(retWord,  "%", w.word)
  retTrans = replace(retTrans, "%", w.translate.get(0))
  ret = TWord(retWord, retTrans, "cardinal")
  ret.map.put("symbol", str(no))
  return ret
end function

function pluralSpanish(sing as string) as string
  ' regular plural forms
  dim as integer lastVocal      = instrrev(sing, any "aeiou")
  dim as integer next2lastVocal = instrrev(sing, any "aeiou", lastVocal-1)
  dim as integer lastAccent     = instrrev(sing, "á"), length = len(sing)
  if lastAccent = 0 then lastAccent = instrrev(sing, "é")
  if lastAccent = 0 then lastAccent = instrrev(sing, "í")
  if lastAccent = 0 then lastAccent = instrrev(sing, "ó")
  if lastAccent = 0 then lastAccent = instrrev(sing, "ú")

  if right(sing, 2) = "ps" andalso lastAccent then return sing
  if lastAccent = length-1 then return sing & "s"
  select case right(sing, 1)
    case "z"
      return left(sing, length-1) & "ces"
    case "s"
      if lastAccent = 0 andalso (next2lastVocal = 0 or _
         (next2lastVocal = lastVocal-1 and instr(mid(sing, next2lastVocal, 2), any "iu") > 0)) then return sing & "es"
      if lastVocal = length - 1 then return sing
      if lastAccent > lastVocal then return modifyTxt(sing, MODIFY_NOACUTE) & "es"
    case "n"
      if lastAccent > lastVocal then return modifyTxt(sing, MODIFY_NOACUTE) & "es"
      if lastAccent orelse next2lastVocal = 0 then return sing & "es"
      return left(sing, next2lastVocal-1) & addAcute(mid(sing, next2lastVocal, 1)) & mid(sing, next2lastVocal+1) & "es"
    case "x"
      if lastAccent then return sing
    case "a", "e", "i", "o", "u", "b", "c", "f", "g", "k", "m", "p", "t"
      return sing & "s"
  end select
  if lastVocal < length-1 andalso lastAccent < length-2 then return sing & "s"
  return sing & "es"
end function

sub speak(txt as string, no as integer = 1, byref ready as integer = 0)
  ' use espeak and mbrolla voice 'es1'
  ' only works if both is installed and ready for access
  mutexlock speakLock
  dim search as integer, t as string = txt
  ready = -1
  mutexunlock speakLock
  search = instr(t, "ll")
  do while search
    t = left(t, search-1) & "y" & mid(t, search+2)
    search = instr(t, "ll")
  loop
  search = instr(t, "rr")
  do while search
    t = left(t, search) & "'" & mid(t, search+1)
    search = instr(t, "rr")
  loop
  shell "espeak -vmb-es" & no & " """ & t & """ -s100"
end sub

sub speakThread(txt as string, no as integer = 1)
  ' calls speak() without waiting for the end of speech
  dim as integer ready = 0, r
  dim as any ptr thread = threadcall speak(txt, no, ready)
  do
    mutexlock speakLock
    r = ready
    mutexunlock speakLock
  loop until r
  threaddetach(thread)
end sub

sub wordInfo(w as TWord)
  dim as integer headX = 120, headY, infoX = 120, infoY = 90, fSize = FONTSIZE.SMALL
  dim as integer line0 = 150, line1 = 190, line2 = 210, row0 = 120, row1 = 250, row2 = 360
  headY = iif(w.groupname = "verb", 20, 50)
  dim as TWord tempWord
  dim as string tempString = w.translate.get(0)
  for i as integer = 1 to w.translate.size - 1
    tempString &= ", " & w.translate.get(i)
  next
  screenlock
  cls
  printFT headX, headY, rtrim(prettyOutput(w, PRETTOUT.OBJ or PRETTOUT.ARTICLE or PRETTOUT.BOTHGEN).word _
                        & " " & w.map.get("obj")) & "  -  " & tempString, font, fSize
  select case w.groupname
    case "noun"
      tempString = "Nomen"
      dim flexion(3) as string, flexBit as integer = 0
      for i as integer = 0 to 3
        flexion(i) = declinSpanish(w, i)
        if len(flexion(i)) then flexBit = bitset(flexBit, i)
      next
      select case flexBit
        case 1
          tempString &= ", maskulin - nur Singular"
        case 2
          tempString &= ", feminin - nur Singular"
        case 3
          tempString &= ", maskulin/feminin - nur Singular"
          if flexion(0) <> flexion(1) then
            printFT row0, line1, "maskulin:", font, fSize
            printFT row0, line2, "feminin:", font, fSize
            printFT row1, line1, "el " & flexion(0), font, fSize
            printFT row1, line2, iif(len(w.map.get("article")), "el ", "la ") & flexion(1), font, fSize
          end if
        case 4
          tempString &= ", maskulin - Pluralwort"
        case 5
          tempString &= ", maskulin"
          printFT row0, line1, "Singular:", font, fSize
          printFT row1, line1, "Plural:", font, fSize
          printFT row0, line2, "el "  & flexion(0), font, fSize
          printFT row1, line2, "los " & flexion(2), font, fSize          
        case 8
          tempString &= ", feminin - Pluralwort"
        case 10
          tempString &= ", feminin"
          printFT row0, line1, "Singular:", font, fSize
          printFT row0, line2, "Plural:", font, fSize
          printFT row1, line1, getArticle(w) & flexion(1), font, fSize
          printFT row1, line2, "las " & flexion(3), font, fSize          
        case 12
          tempString &= " - Pluralwort"
          printFT row0, line1, "maskulin:", font, fSize
          printFT row0, line2, "feminin:", font, fSize
          printFT row1, line1, "los " & flexion(2), font, fSize
          printFT row1, line2, "las " & flexion(3), font, fSize
        case 15
          tempString &= ", maskulin/feminin"
          printFT row0, line1, "maskulin:", font, fSize
          printFT row0, line2, "feminin:", font, fSize
          printFT row1, line0, "Singular:", font, fSize
          printFT row2, line0, "Plural:", font, fSize
          printFT row1, line1, "el "  & flexion(0), font, fSize
          printFT row1, line2, "la "  & flexion(1), font, fSize
          printFT row2, line1, "los " & flexion(2), font, fSize
          printFT row2, line2, "las " & flexion(3), font, fSize
        case else
          tempString &= " <" & flexBit & ">"
      end select
      printFT infoX, infoY, tempString, font, fSize
    case "adjective"
      tempString = "Adjektiv"
      dim flexion(3) as string, flexBit as integer = 0
      for i as integer = 0 to 3
        flexion(i) = declinSpanish(w, i)
      next
      if flexion(0) = flexion(1) andalso flexion(0) = flexion(2) then
        tempString &= " - unveränderlich"
      else
        ' Tabelle
      end if
      printFT infoX, infoY, tempString, font, fSize
    case "adverb"
      printFT infoX, infoY, "Adverb", font, fSize
    case "pronoun"
      printFT infoX, infoY, "Pronomen", font, fSize
    case "interrogative"
      printFT infoX, infoY, "Interrogativ", font, fSize
    case "preposition"
      printFT infoX, infoY, "Präposition", font, fSize
    case "conjunction"
      printFT infoX, infoY, "Konjunktion", font, fSize
    case "particle"
      printFT infoX, infoY, "Partikel", font, fSize
    case "phrase"
      printFT infoX, infoY, "feststehender Ausdruck", font, fSize
    case "cardinal"
      printFT infoX, infoY, "Kardinalzahl", font, fSize
    case "comparison"
      printFT infoX, infoY, "Vergleich", font, fSize
    case "verb" 
      dim as integer conjX(3) = { 120, 280, 440, 600 }, conjY(3) = { 50, 190, 330, 470 }
      dim as integer conjPersonX = 30, conjLines = 15, conjSize = FONTSIZE.TINY
      dim as string person(1 to 6) = { "yo", "tú", "él/ella", "nosotr.", "vosotr.", "ellos/-as" }
      dim as string headline(15) = { _
          "Presente Ind.", "Imperfecto Ind.", "Indefinido", "Condicional", _
          "Futuro Simple Ind.", "Futuro Perfecto", "Perfecto", "Pluscuamperfecto", _
          "Presente Subj.", "Imperfecto Subj. I", "Imperfecto Subj. II", "Condicional Perfecto", _
          "Futuro Subj.",	"Futuro Perf.Subj.", "Perfecto Anterior", "Imperativo" }
      dim as integer tense(15) = { _
          INFLECT.PRES, INFLECT.PRETIMP, INFLECT.PRETIND, INFLECT.COND, _
          INFLECT.FUTSIMP, INFLECT.FUTCOMP, INFLECT.PERF, INFLECT.PQPERF, _
          INFLECT.PRES or INFLECT.SUBJ, INFLECT.PRETIMP or INFLECT.SUBJ, _
          INFLECT.PRETIND or INFLECT.SUBJ, INFLECT.CONDPERF, INFLECT.FUTSIMP or INFLECT.SUBJ, _
          INFLECT.FUTCOmP or INFLECT.SUBJ, INFLECT.PERFANT, INFLECT.IMPERAT }
      for col as integer = 0 to 3
        for p as integer = 1 to 6
          printFT conjPersonX, conjY(col) + (p+.5)*conjLines, person(p), font, conjSize
        next
      next
      for x as integer = 0 to 3
        for y as integer = 0 to 3
          printFT conjX(x), conjY(y), headline(y*4+x), font, conjSize+2
          for p as integer = 1 to 3
            if p = 1 and tense(y*4+x) = INFLECT.IMPERAT then
              printFT conjX(x), conjY(y)+(p+.5)*conjLines, "-----", font, conjSize
            else
              tempWord = conjugSpanish(w, tense(y*4+x) or p)
              printFT conjX(x), conjY(y)+(p+.5)*conjLines, tempWord.word, _
                  font, conjSize, iif(tempWord.mark, &hff0000, 0)
            end if
            tempWord = conjugSpanish(w, tense(y*4+x) or (p+INFLECT.PLURAL))
            printFT conjX(x), conjY(y)+(p+3.5)*conjLines, tempWord.word, _
                font, conjSize, iif(tempWord.mark, &hff0000, 0)
          next
        next
      next
    case else
      printFT infoX, infoY, "keine weiteren Informationen verfügbar", font, fSize
  end select
  screenunlock
  getkey
end sub
