' module: print all words in a single HTML page or in a page for each lession

namespace EXPORTLIST
  const usePages = 1, useColor = 2, useScript = 4, useSortL = 8, useSortT = 16, useSortA = 32, useSortM = 64
  declare function callback(exercise as TExercise) as long
  declare function outp(x as integer, y as integer, x2 as integer, h as integer, _
                        word as TWord, opt as integer) as integer
  declare sub printHeader(filenr as integer, title as string = "")
  declare sub printFooter(filenr as integer)
  declare function settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  declare function toggle(gui as TGui, button as TGuiElement, state as integer, toggleVal as integer) as integer

  type TGroup
    as string        id
    as integer       count
    as mdList(TWord) list
    declare constructor
    declare constructor(id_ as string)
    declare destructor
    declare operator cast as string
  end type
  constructor TGroup
  end constructor
  constructor TGroup(id_ as string)
    id = id_
  end constructor
  destructor TGroup
    list.clear
  end destructor
  operator TGroup.cast as string
    return id
  end operator
  operator =(g1 as TGroup, g2 as tGroup) as integer
    return g1.id = g2.id
  end operator
  mdListDeclare(TGroup)
end namespace

globalOptions.add(TExercise("export", "Wortliste exportieren", "full", 0, _
                  @EXPORTLIST.callback, 0, @EXPORTLIST.settings))

function EXPORTLIST.callback(exercise as TExercise) as long
  dim as integer exportType = exercise.settingString[0]
  if exportType and usePages andalso (useSortL or useSortT) = 0 then exportType and= not usePages
  dim as string path = "export/" & LANGUAGE & "/", filename = "Woerter-" & LANGUAGE, trigger = "export.sh"

  dim as integer ret, filenumber, tempInt, span = 0, groupCount = 0
  dim as string curID, tempString, obj, article, genus, fem, suffix, sortWord
  dim lp as TListPoint, w as TWord
  dim as TGroup group(exercise.currentList.list.size)

  ' prepare word list
  foreach(TListPoint, lp, exercise.currentList.list)
    w = globallist.get(lp.id)
    if len(w.word) = 0 then continue while
    ' search corresponding group or place for a new group
    if exportType and useSortL then curID = str(w.lession) else curID = ""
    for t as integer = 0 to iif(exportType and useSortT, w.tag.size-1, 0)
      if exportType and useSortT then curID = w.tag.get(t)
      for i as integer = 0 to groupCount - 1
        if curID = group(i).id then
          ' add word to existing group
          sortWord = modifyTxt(w.word, MODIFY_SORTABLE)
          group(i).count += 1
          if exportType and useSortA then
            for k as integer = 0 to group(i).count-1
              ' looking for right place to insert into current group
              if sortWord < modifyTxt(group(i).list.get(k).word, MODIFY_SORTABLE) then
                group(i).list.add(k, w)
                group(i).count += 1
                continue while
              end if
            next
            ' no place found - append to current group
            group(i).list.add(w)
            group(i).count += 1
            continue while
          elseif exportType and useSortM then
            ' choose random place
            tempInt = int(rnd*group(i).count)
            if tempInt = group(i).count-1 then
              group(i).list.add(w)
            else
              group(i).list.add(tempInt, w)
            end if
            continue while
          else
            ' append
            group(i).list.add(w)
            group(i).count += 1
            continue while
          end if
        end if
      next
    next
    ' build new curGroup and append to group
    if (exportType and useSortT) = 0 orelse w.tag.size > 0 then
      group(groupCount) = TGroup(curID)
      group(groupCount).list.add(w)
      group(groupCount).count = 1
      groupCount += 1
    end if
  nexteach

  ' write export file(s)  
  if (exportType and usePages) = 0 then
    ' export to single file
    filenumber = freefile
    open exepath & "/" & path & filename & ".htm" for output as #filenumber
    printHeader filenumber
  end if
  for i as integer = 0 to groupCount-1
    if group(i).count = 0 then continue for
    if exportType and usePages then
      ' export to different files
      filenumber = freefile
      open exepath & "/" & path & filename & "_" & modifyTxt(group(i).id, MODIFY_ALPHANUM) & ".htm" for output as #filenumber
      if exportType and useSortL then
        tempInt = cint(group(i).id)
        printHeader filenumber, "Lektion " & iif(tempInt = 0, "0", tempInt\100 & "." & tempInt mod 100)
      elseif exportType and useSortL then
        printHeader filenumber, group(i).id
      else
        printHeader filenumber, group(i).id
      end if
    end if
    print #filenumber, "<h1 id='h" & modifyTxt(group(i).id, MODIFY_ALPHANUM) & "' class='newpage'>";
    if exportType and useSortL then
      tempInt = cint(group(i).id)
      print #filenumber, "Lektion " & iif(tempInt = 0, "0", tempInt\100 & "." & tempInt mod 100) & "</h1>"
    elseif exportType and useSortT then
      print #filenumber, group(i).id & "</h1>"
    else
      print #filenumber, "Wörterliste</h1>"
    end if
    print #filenumber, "<table>"
    foreach(TWord, w, group(i).list)
      tempstring = ""
      suffix = iif(len(w.map.get("article")), " <i>(f)</i>", "")
      span = 0  
      print #filenumber, "  <tr>"
      print #filenumber, "    <td>";
      fem = w.map.get("f")
      genus = w.map.get("genus")
      article = getArticle(w, -1)
      if len(genus) andalso len(fem) = 0 andalso exportType and useColor then
        span = -1
        print #filenumber, "<span style='color:" & iif(genus = "m", "blue", "red") & ";'>" & article;
      else
        print #filenumber, article;
      end if
      if len(fem) then
        if left(fem, len(fem)-1) = left(w.word, len(w.word)-1) or left(fem, len(fem)-1) = w.word then
          tempString = "/-" & right(fem, 1)
        elseif left(fem, len(fem)-2) = left(w.word, len(w.word)-2) then
          tempString = "/-" & right(fem, 2)
        else
          tempString = "/" & fem
        end if
      end if
      obj = w.map.get("obj")
      if len(obj) then obj = " <i>" & obj & "</i>"
      print #filenumber, w.word & tempString & obj & iif(span, "</span>", "") & suffix & "&nbsp;&nbsp;";
      tempString = w.map.get("irr")
      if len(tempString) then print #filenumber, "<br />&nbsp;&nbsp;&nbsp;<i>(" & tempString & ")</i>";
      print #filenumber, "</td>"
      print #filenumber, "    <td>" & w.translate.get(0);
      for i as integer = 1 to w.translate.size-1
        print #filenumber, ", " & w.translate.get(i);
      next
      print #filenumber, "</td>"
      print #filenumber, "  </tr>"
    nexteach
    print #filenumber, "</table>"
    if exportType and usePages then
      ' close file (export to different files)
      printFooter filenumber
      close #filenumber
    end if
  next
  if (exportType and usePages) = 0 then
    ' close file (export to single file)
    printFooter(filenumber)
    close filenumber
  end if

  ' execute extern script
  if exportType and useScript then
    messageWait "Skript wird ausgeführt - bitte warten"
    if shell(exepath & "/" & path & trigger) then
      messageWait
      ret = message("FEHLER: Das Skript konnte nicht korrekt ausgeführt werden!",, 1)
    else
      messageWait
      ret = message("Das Skript wurde ausgeführt.",, 1)
    end if
  else
    ret = message("Die Export-Datei wurde erstellt und liegt im Ordner " & path,, 1)
  end if
  return iif(ret = RETVAR.windowClosed, ret, 0)
end function

sub EXPORTLIST.printHeader(filenr as integer, title as string = "")
  print #filenr, "<!DOCTYPE html>"
  print #filenr, "<html xmlns='http://www.w3.org/1999/xhtml'>"
  print #filenr, "<head>"
  print #filenr, "  <title>Spanisch-W&ouml;rter";
  if len(title) then
    print #filenr, ", "; title;
  end if
  print #filenr, "</title>"
  print #filenr, "  <meta http-equiv='content-type' content='text/html;charset=utf-8' />"
  print #filenr, "  <style type='text/css'>"
  print #filenr, "    td { vertical-align:top; }"
  print #filenr, "  </style>"
  print #filenr, "</head>"
  print #filenr, "<body>"
  print #filenr, "<div>"
end sub

sub EXPORTLIST.printFooter(filenr as integer)
  print #filenr, "</div>"
  print #filenr, "</body>"
  print #filenr, "</html>"
end sub

function EXPORTLIST.settings(exercise as TExercise, optional as long = SETTING_OPTIONS.NONE) as long
  enum
    posPages = 1, posColor, posScript, posSortL, posSortT, posSortA, posSortM, posWords, posGo, posExit
  end enum
  if len(exercise.settingString) < 4 then exercise.settingString = mkl(9)
  dim as integer buttonState = cvl(left(exercise.settingString, 4)), ret, mx, my, mb, tempInt
  dim as string wordString = mid(exercise.settingString, 5)
  tempInt = chooseWords(exercise, wordString, SETTING_OPTIONS.SKIPDIALOG)

  ' if this was the initial call, nothing else is to do
  if optional and SETTING_OPTIONS.SKIPDIALOG then return tempInt
  
  dim as TGui settingGui
  dim as TGuiEvent event
  dim as string key, scriptname
  #ifdef __FB_PCOS__
    scriptname = "export/" & LANGUAGE & "/export.bat"
  #else
    scriptname = "export/" & LANGUAGE & "/export.sh"
  #endif
  dim as integer stX = 95, butW = 130, butH = 30, dist = 30, row1 = 100, row2 = 300, row3 = 500
  
  dim as TGuiElement butPages  = TGuiElement(posPages,  stX,               row1, butW, butH, "mehrere Seiten",    "p",, "P"), _
                    butColor  = TGuiElement(posColor,  stX+(butW+dist),   row1, butW, butH, "farbige Ausgabe",   "c",, "C"), _
                    butScript = TGuiElement(posScript, stX+(butW+dist)*2, row1, butW, butH, "Skript aufrufen",   "s",, "S"), _
                    butSortL  = TGuiElement(posSortL,  stX,               row2, butW, butH, "nach Lektion",      "l",, "L"), _
                    butSortT  = TGuiElement(posSortT,  stX+(butW+dist),   row2, butW, butH, "nach Tags",         "t",, "T"), _
                    butSortA  = TGuiElement(posSortA,  stX+(butW+dist)*2, row2, butW, butH, "alphabetisch",      "a",, "A"), _
                    butSortM  = TGuiElement(posSortM,  stX+(butW+dist)*3, row2, butW, butH, "zufällig gemischt", "m",, "M"), _
                    butWords  = TGuiElement(posWords,  stX,               row3, butW, butH, "Wörter auswählen",  chr(9),,  "Tab"), _
                    butGo     = TGuiElement(posGo,     stX+(butW+dist)*2, row3, butW, butH, "EXPORTIEREN",       chr(13),, "Enter"), _
                    butExit   = TGuiElement(posExit,   stX+(butW+dist)*3, row3, butW, butH, "zurück",            chr(27),, "Esc")
  ' activate buttons
  if buttonState and usePages  then butPages.state  = GuiState.activated
  if buttonState and useColor  then butColor.state  = GuiState.activated
  if buttonState and useScript then butScript.state = GuiState.activated
  if buttonState and useSortL  then
    butSortL.state  = GuiState.activated
    buttonState and= not useSortT
  else
    if buttonState and useSortT then butSortT.state  = GuiState.activated
  end if
  if buttonState and useSortA  then
    butSortA.state  = GuiState.activated
    buttonState and= not useSortM
  else
    if buttonState and useSortM then butSortM.state  = GuiState.activated
  end if
  settingGui.addElement butPages
  settingGui.addElement butColor
  settingGui.addElement butScript
  settingGui.addElement butSortL
  settingGui.addElement butSortT
  settingGui.addElement butSortA
  settingGui.addElement butSortM
  settingGui.addElement butWords
  settingGui.addElement butGo
  settingGui.addElement butExit
  screenlock
  cls
  printFT stX, row1-40, "Ausgabe-Einstellungen", font, 18
  printFT stX, row1+butH+45, "Die mehrseitige Ausgabe funktioniert nur, wenn zugleich nach Lektionen sortiert wird.", font, 14
  printFT stX, row1+butH+65, """Skript aufrufen"" exportiert normal und ruft dann das Skript " & scriptname & " auf.", font, 14
  printFT stX, row2-40, "Sortierung", font, 18
  printFT stX, row2+butH+45, "Wenn Sie nach Lektionen sortieren, stehen alle Wörter derselben Lektion beieinander.", font, 14
  printFT stX, row2+butH+65, "Unabhängig davon können Sie zusätzlich eine alphabetische oder zufällige Reihenfolge wählen.", font, 14
  printFT stX, row2+butH+85, "Wenn Sie weder alphabetisch sortieren noch zufällig mischen, erfolgt die Sortierung nach ID.", font, 14
  settingGui.show
  screenunlock
  do
    key = inkey
    if key = chr(255, 107) then ret = RETVAR.windowClosed : exit do
    event = settingGui.getEvent(key)
    select case event.id
      case posPages
        buttonState = toggle(settingGui, butPages,  buttonState, usePages)
      case posColor
        buttonState = toggle(settingGui, butColor,  buttonState, useColor)
      case posScript
        buttonState = toggle(settingGui, butScript, buttonState, useScript)
      case posSortL
        buttonState = toggle(settingGui, butSortL,  buttonState, useSortL)
        if (buttonState and (useSortL or useSortT)) = (useSortL or useSortT) then
          buttonState = toggle(settingGui, butSortT, buttonState, useSortT)
        end if
      case posSortT
        buttonState = toggle(settingGui, butSortT,  buttonState, useSortT)
        if (buttonState and (useSortL or useSortT)) = (useSortL or useSortT) then
          buttonState = toggle(settingGui, butSortL, buttonState, useSortL)
        end if
      case posSortA
        buttonState = toggle(settingGui, butSortA,  buttonState, useSortA)
        if (buttonState and (useSortA or useSortM)) = (useSortA or useSortM) then
          buttonState = toggle(settingGui, butSortM, buttonState, useSortM)
        end if
      case posSortM
        buttonState = toggle(settingGui, butSortM,  buttonState, useSortM)
        if (buttonState and (useSortA or useSortM)) = (useSortA or useSortM) then
          buttonState = toggle(settingGui, butSortA, buttonState, useSortA)
        end if
      case posWords
        dim as FB.Image ptr tempImg = imagecreate(SCREEN_X, SCREEN_Y)
        get (0, 0)-(SCREEN_X-1, SCREEN_Y-1), tempImg
        ret = chooseWords(exercise, wordString)
        put (0, 0), tempImg, pset
        imagedestroy tempImg
        if ret = RETVAR.windowClosed then exit do
      case posGo
        ret = callback(exercise)
        if ret = RETVAR.windowClosed then exit do
      case posExit
        ret = RETVAR.ok
        exit do
    end select
    sleep 1
    if event.id then
      do
        getmouse mx, my,, mb
        sleep 1
      loop until mb < 1
    end if
  loop
  exercise.settingString = mkl(buttonState) & wordString
  return ret
end function

function EXPORTLIST.toggle(gui as TGui, button as TGuiElement, state as integer, toggleVal as integer) as integer
  button.state = iif(button.state = GuiState.activated, GuiState.active, GuiState.activated)
  gui.setElement button
  button.show
  return state xor toggleVal
end function
