const VERSION_MAJOR = 2, VERSION_MINOR = 3, VERSION_PATCH = 170714

#include once "include/shared.bi"
#ifdef __FB_UNIX__
  dim as string pre = "liblengua_", ending = ".so"
#elseif defined(__FB_PCOS__)
  dim as string pre = "lengua_", ending = ".dll"
#else
  #error "CRITICAL ERROR: system not supported"
#endif

const posExit = -1
const ALLOW_EDITS = 1

declare sub writeOptionFile(options as integer)

dim as TModule mainModule = TModule("main", "Lengua", VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH)
dim as any ptr library
dim getInfo as sub(module as TModule)
dim init    as sub(mainModule as TModule ptr)
dim as string key
dim as integer countLang = 0, startModule = -1, options = 0
redim as TModule module(99)

module(countLang).libname = dir(exepath & "/modules/" & pre & "*" & ending)
do while len(module(countLang).libname)
  library = dylibload(exepath & "/modules/" & module(countLang).libname)
  if library = 0 then
    module(countLang).libname = dir
    continue do
  end if
  getInfo = dylibsymbol(library, "GETINFO")
  if getInfo = 0 then
    module(countLang).libname = dir
    dylibfree library
    continue do
  end if
  getInfo(module(countLang))
  dylibfree library
  if len(module(countLang).libShort) = 0 orelse module(countLang).api <> mainModule.api then
    module(countLang).libname = dir
    continue do
  end if
  countLang += 1
  if countLang > ubound(module) then redim preserve module(countLang+10)
  module(countLang).libname = dir
loop
if countLang > 50 then countLang = 50

' CHECK ARGUMENTS
for i as integer = 1 to __FB_ARGC__ - 1
  if command(i) = "-h" or command(i) = "--help" then
    print "Lengua v" & VERSION_MAJOR & "." & VERSION_MINOR & "." & VERSION_PATCH
    print "API version: " & mainModule.api shr 16 & "." & mainModule.api mod 16
    print "Read doc/aboutLengua.htm for more information."
    print "Usage: lengua [ OPTIONS ]"
    print
    print "Options supported by lengua:"
    print "  -h or --help      Print a help message"
    print "  -l or --list      List available modules"
    print "  -m=MODULENAME     Start module MODULENAME (if possible)"
    print "  -e or --edit      Allow edits inside of the modules"
    print
    end
  elseif command(i) = "-l" or command(i) = "--list" then
    print "Available modules:"
    for i as integer = 0 to countLang-1
      print module(i).libShort; tab(12); module(i).title; tab(32);
      print " (v"; module(i).major & "." & module(i).minor & "." & module(i).patch & ")"
    next
    print
    end
  elseif command(i) = "-e" or command(i) = "--edit" then
    options or= ALLOW_EDITS
  elseif left(command(i), 3) = "-m=" then
    for k as integer = 0 to countLang-1
      if len(module(k).libShort) andalso module(k).libShort = mid(command(i), 4) then
        startModule = k
        exit for
      end if
    next
    if startModule < 0 then
      print "Can't load module '" & mid(command(i), 4)  & "'!"
      print "Maybe the module isn't installed or uses the wrong API version."
      print
      end
    end if
  else
    print "Invalide option " & command(i)
    print "Use 'lengua -h' to see the available options."
    print
    end
  end if
next

if countLang = 1 then startModule = 0
if startModule >= 0 then
  library = dylibload(exepath & "/modules/" & module(startModule).libname)
  init = dylibsymbol(library, "INIT")
  writeOptionFile options
  init(@mainModule)
  dylibfree library
  kill OPTIONS_FILENAME
  end
end if

' INITIALIZE FONT
#include once "include/font.bi"
FT_Var.ErrorMsg = FT_Init_FreeType(@FT_Var.Library)
if FT_Var.ErrorMsg then print "Could not load library"
dim shared as FT_Face font
font = getFont(exepath & "/font.ttf")
if font = 0 then print "couldn't find font.ttf" : getkey : end

' CREATE GUI
type TExercise
  ' dummy exercise type (for TGui.bi)
  as function(as TExercise) as string info
end type
#include once "include/TGui.bi"
dim as TGui gui
dim as TGuiEvent event
dim as TGuiElement button(countLang-1)
for i as integer = 0 to countLang-1
  button(i) = TGuiElement(GuiElementType.button, i+1, _
              (i mod 5)*120+30, (i\5)*50+80, 100, 30, module(i).title)
  gui.addElement button(i)
next
dim as TGuiElement buttonOk = TGuiElement(GuiElementType.button, posExit, _
                              510, ((countLang+4)\5)*50+100, 100, 30, "Ende", chr(255, 107))
gui.addElement buttonOk
screenres 660, 50*((countLang+4)\5) + 160, 32
windowtitle mainModule.title & " v" & mainModule.major & "." & mainModule.minor & "." & mainModule.patch

color rgb(0, 0, 0), rgb(255, 255, 255)
cls
screenlock
if countLang = 0 then
  centerFT 300, 40, "Es steht kein passendes Modul zur Verfügung.", font, FONTSIZE.HEADLINE
else
  centerFT 300, 40, "Bitte das gewünschte Lern-Modul wählen:", font, FONTSIZE.HEADLINE
end if
gui.show
screenunlock

' MAIN PROGRAM
do
  key = inkey
  event = gui.getEvent(key)
  if event.id > 0 then
    screenlock
    cls
    centerFT 300, 40, "Bitte das gewünschte Lern-Modul wählen:", font, FONTSIZE.HEADLINE, rgb(127, 127, 127)
    button(event.id-1).state = GuiState.activated
    gui.show GuiState.deactivated
    button(event.id-1).state = GuiState.activated or GuiState.deactivated
    button(event.id-1).show
    button(event.id-1).state = GuiState.active
    screenunlock
    library = dylibload(exepath & "/modules/" & module(event.id-1).libname)
    init = dylibsymbol(library, "INIT")
    writeOptionFile options
    init(@mainModule)
    dylibfree library
    kill OPTIONS_FILENAME
    screenlock
    cls
    centerFT 300, 40, "Bitte das gewünschte Lern-Modul wählen:", font, FONTSIZE.HEADLINE
    gui.show
    screenunlock
  end if
  sleep 1
loop until event.id = posExit

sub writeOptionFile(options as integer)
  dim as integer fileNumber = freefile
  open OPTIONS_FILENAME for output as #fileNumber
  if options and AllOW_EDITS then
    print #fileNumber, "EDITS=ALLOWED"
  end if
  close fileNumber
end sub
